--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- Mock Turtle
-- https://gitlab.cern.ch/coht/mockturtle
--------------------------------------------------------------------------------
--
-- unit name:   mt_config_rom
--
-- description: A ROM to hold the various configuration values of the current
-- MockTurtle implementation.
--
--------------------------------------------------------------------------------
-- Copyright (c) 2018 CERN
--------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
--------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.mock_turtle_pkg.all;
use work.mt_mqueue_pkg.all;

use work.wishbone_pkg.all;

entity mt_config_rom is
  generic(
    -- Message Queue and CPU configuration
    g_config            : t_mock_turtle_config := c_default_mock_turtle_config;
    -- When true, the Remote Message Queue is implemented.
    g_with_rmq          : boolean              := TRUE;
    -- Frequency of clk_sys_i, in Hz
    g_system_clock_freq : integer              := 62500000;
    -- Enables/disables WR support
    g_with_white_rabbit : boolean              := FALSE);
  port
    (
      clk_i    : in  std_logic;
      rst_n_i  : in  std_logic;
      slave1_i : in  t_wishbone_slave_in;
      slave1_o : out t_wishbone_slave_out;
      slave2_i : in  t_wishbone_slave_in;
      slave2_o : out t_wishbone_slave_out);
end entity;

architecture rtl of mt_config_rom is

  constant c_rom_size    : integer := 256;
  type t_rom_array is array(0 to c_rom_size-1) of std_logic_vector(31 downto 0);
  constant c_mt_revision : integer := 20161208;

  function f_char_to_slv (c : character) return std_logic_vector is
  begin
    return std_logic_vector(to_unsigned(character'pos(c), 8));
  end f_char_to_slv;

  function f_bool_to_sl (x : boolean) return std_logic is
  begin
    if x then
      return '1';
    else
      return '0';
    end if;
  end f_bool_to_sl;

  function f_int_to_slv(x : integer) return std_logic_vector is
  begin
    return std_logic_vector(to_unsigned(x, 32));
  end f_int_to_slv;

  procedure f_make_mq_config (
    cfg    : out t_rom_array;
    offset :     integer;
    mq     :     t_mt_mqueue_config
    ) is
  begin
    cfg(offset)   := f_int_to_slv(mq.in_slot_count);
    cfg(offset+1) := f_int_to_slv(mq.out_slot_count);
    for i in 0 to 15 loop
      cfg(offset + 2 * i + 2)     := f_int_to_slv(mq.in_slot_config(i).width);
      cfg(offset + 2 * i + 2 + 1) := f_int_to_slv(mq.in_slot_config(i).entries);
    end loop;
    for i in 0 to 15 loop
      cfg(offset + 32 + 2 * i + 2)     := f_int_to_slv(mq.out_slot_config(i).width);
      cfg(offset + 32 + 2 * i + 2 + 1) := f_int_to_slv(mq.out_slot_config(i).entries);
    end loop;

  end f_make_mq_config;

  function f_initialize_rom return t_rom_array is
    variable cfg : t_rom_array;
  begin

    --init to zero
    cfg := (others => (others => '0'));

    cfg(0)    := f_char_to_slv('T') & f_char_to_slv('R') & f_char_to_slv('T') & f_char_to_slv('L');
    cfg(1)    := f_int_to_slv(c_mt_revision);
    cfg(3)    := f_int_to_slv(g_system_clock_freq);
    cfg(4)(0) := f_bool_to_sl(g_with_white_rabbit);
    cfg(4)(1) := f_bool_to_sl(g_with_rmq);
    cfg(5)    := g_config.app_id;
    cfg(6)    := f_int_to_slv(g_config.cpu_count);
    cfg(7)    := f_int_to_slv(g_config.shared_mem_size);

    for i in 0 to 7 loop
      if (i < g_config.cpu_count) then
        cfg(8+i) := f_int_to_slv(g_config.cpu_memsizes(i));
      else
        cfg(8+i) := (others => '0');
      end if;
    end loop;

    -- word 64
    f_make_mq_config(cfg, 64, g_config.hmq_config);
    f_make_mq_config(cfg, 64 + 64 + 2, g_config.rmq_config);

    return cfg;

  end f_initialize_rom;

  constant c_rom_data : t_rom_array := f_initialize_rom;

  signal wb1_ack, wb2_ack : std_logic := '0';

begin

  p_ack : process(clk_i)
  begin
    if rising_edge(clk_i) then
      if rst_n_i = '0' then
        wb1_ack <= '0';
        wb2_ack <= '0';
      else
        wb1_ack <= slave1_i.cyc and slave1_i.stb;
        wb2_ack <= slave2_i.cyc and slave2_i.stb;
      end if;
    end if;
  end process p_ack;

  slave1_o.ack   <= wb1_ack;
  slave1_o.dat   <= c_rom_data(to_integer(unsigned(slave1_i.adr(9 downto 2))));
  slave1_o.stall <= '0';
  slave1_o.err   <= '0';
  slave1_o.rty   <= '0';
  slave1_o.int   <= '0';

  slave2_o.ack   <= wb2_ack;
  slave2_o.dat   <= c_rom_data(to_integer(unsigned(slave2_i.adr(9 downto 2))));
  slave2_o.stall <= '0';
  slave2_o.err   <= '0';
  slave2_o.rty   <= '0';
  slave2_o.int   <= '0';

end rtl;
