--------------------------------------------------------------------------------
-- CERN BE-CO-HT
-- Mock Turtle
-- https://gitlab.cern.ch/coht/mockturtle
--------------------------------------------------------------------------------
--
-- unit name:   svec_mt_demo
--
-- description: Top entity for Mock Turtle SVEC demo.
--
-- This demo instantiates a Mock Turtle CPU with two cores, each connected to
-- a WB GPIO peripheral. Reading and writing to the GPIOs controls the LEDs
-- and LEMO I/O on the front panel of the SVEC.
--
--------------------------------------------------------------------------------
-- Copyright (c) 2018 CERN
--------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
--------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.gencores_pkg.all;
use work.wishbone_pkg.all;
use work.vme64x_pkg.all;
use work.mock_turtle_pkg.all;
use work.mt_mqueue_pkg.all;
use work.synthesis_descriptor.all;

library unisim;
use unisim.vcomponents.all;

entity svec_mt_demo is
  generic (
    -- Bypass VME core, useful only in simulation
    g_sim_bypass_vme : boolean := FALSE);
  port (
    -- resets and clocks
    rst_n_a_i           : in std_logic;
    clk_125m_pllref_p_i : in std_logic;
    clk_125m_pllref_n_i : in std_logic;

    -- SVEC Front panel LEDs and GPIO
    fp_led_line_oen_o : out   std_logic_vector(1 downto 0);
    fp_led_line_o     : out   std_logic_vector(1 downto 0);
    fp_led_column_o   : out   std_logic_vector(3 downto 0);
    fp_gpio1_a2b_o    : out   std_logic;
    fp_gpio2_a2b_o    : out   std_logic;
    fp_gpio34_a2b_o   : out   std_logic;
    fp_gpio1_b        : inout std_logic;
    fp_gpio2_b        : inout std_logic;
    fp_gpio3_b        : inout std_logic;
    fp_gpio4_b        : inout std_logic;

    -- Bypass VME core, useful only in simulation
    -- synthesis translate_off
    sim_wb_i : in  t_wishbone_slave_in := cc_dummy_slave_in;
    sim_wb_o : out t_wishbone_slave_out;
    -- synthesis translate_on

    -- VME Interface pins
    VME_AS_n_i      : in    std_logic;
    VME_RST_n_i     : in    std_logic;
    VME_WRITE_n_i   : in    std_logic;
    VME_AM_i        : in    std_logic_vector(5 downto 0);
    VME_DS_n_i      : in    std_logic_vector(1 downto 0);
    VME_GA_i        : in    std_logic_vector(5 downto 0);
    VME_BERR_o      : inout std_logic;
    VME_DTACK_n_o   : inout std_logic;
    VME_RETRY_n_o   : out   std_logic;
    VME_RETRY_OE_o  : out   std_logic;
    VME_LWORD_n_b   : inout std_logic;
    VME_ADDR_b      : inout std_logic_vector(31 downto 1);
    VME_DATA_b      : inout std_logic_vector(31 downto 0);
    VME_BBSY_n_i    : in    std_logic;
    VME_IRQ_n_o     : out   std_logic_vector(6 downto 0);
    VME_IACK_n_i    : in    std_logic;
    VME_IACKIN_n_i  : in    std_logic;
    VME_IACKOUT_n_o : out   std_logic;
    VME_DTACK_OE_o  : inout std_logic;
    VME_DATA_DIR_o  : inout std_logic;
    VME_DATA_OE_N_o : inout std_logic;
    VME_ADDR_DIR_o  : inout std_logic;
    VME_ADDR_OE_N_o : inout std_logic);

end svec_mt_demo;

architecture arch of svec_mt_demo is

  -- Host Message Queue configuration
  -- Each CPU has three queues assigned:
  -- 1) a small incoming queue (8 entries x 32 words) for sending host
  -- to CPU commands
  -- 2) a large outgoing queue (128 entries x 16 words) for optional event
  -- logging (asynchronous)
  -- 3) a small outgoing queue (8 entries x 128 words) for receiving
  -- command replies (CPU to host)
  constant c_hmq_config : t_mt_mqueue_config := (
    out_slot_count  => 4,
    out_slot_config => (
      0             => (width => 128, entries => 8),  -- control CPU 0 (to host)
      1             => (width => 128, entries => 8),  -- control CPU 1 (to host)
      2             => (width => 16, entries => 128),  -- log CPU 0
      3             => (width => 16, entries => 128),  -- log CPU 1
      others        => (0, 0)),
    in_slot_count   => 2,
    in_slot_config  => (
      0             => (width => 32, entries => 8),  -- control CPU 0 (from host)
      1             => (width => 32, entries => 8),  -- control CPU 1 (from host)
      others        => (0, 0)));

  constant c_mt_config : t_mock_turtle_config := (
    app_id          => x"d330d330",
    cpu_count       => 2,
    cpu_memsizes    => (32768, 32768, 0, 0, 0, 0, 0, 0),
    hmq_config      => c_hmq_config,
    rmq_config      => c_dummy_mt_mqueue_config,
    shared_mem_size => 8192,
    tpu_enable      => FALSE,
    tpu_channels    => 1,
    tpu_buffer_size => 1024);

  constant c_NUM_WB_MASTERS : integer := 2;
  constant c_NUM_WB_SLAVES  : integer := 1;

  constant c_MASTER_VME : integer := 0;

  constant c_SLAVE_VIC      : integer := 0;
  constant c_SLAVE_MT       : integer := 1;
  constant c_DESC_SYNTHESIS : integer := 2;
  constant c_DESC_REPO_URL  : integer := 3;

  constant c_INTERCONNECT_LAYOUT : t_sdb_record_array(c_NUM_WB_MASTERS + 1 downto 0) := (
    c_SLAVE_VIC      => f_sdb_embed_device(c_xwb_vic_sdb, x"00002000"),
    c_SLAVE_MT       => f_sdb_embed_device(c_mock_turtle_sdb, x"00020000"),
    c_DESC_SYNTHESIS => f_sdb_embed_synthesis(c_sdb_synthesis_info),
    c_DESC_REPO_URL  => f_sdb_embed_repo_url(c_sdb_repo_url));

  constant c_SDB_ADDRESS : t_wishbone_address := x"00000000";

  constant c_VIC_VECTOR_TABLE : t_wishbone_address_array(0 to 1) := (
    0 => x"00020000",   -- MT Mqueue interrupt
    1 => x"00020001");  -- MT Debug Msg interrupt

  signal cnx_master_out : t_wishbone_master_out_array(c_NUM_WB_MASTERS-1 downto 0);
  signal cnx_master_in  : t_wishbone_master_in_array(c_NUM_WB_MASTERS-1 downto 0);

  signal cnx_slave_out : t_wishbone_slave_out_array(c_NUM_WB_SLAVES-1 downto 0);
  signal cnx_slave_in  : t_wishbone_slave_in_array(c_NUM_WB_SLAVES-1 downto 0);

  signal clk_125m_pllref : std_logic;
  signal clk_sys         : std_logic;

  signal rst_n_sys : std_logic;

  signal pllout_clk_sys       : std_logic;
  signal pllout_clk_fb_pllref : std_logic;

  signal dp_wb_out : t_wishbone_master_out_array(1 downto 0);
  signal dp_wb_in  : t_wishbone_master_in_array(1 downto 0);

  signal cpu0_gpio_oen, cpu1_gpio_oen : std_logic_vector(23 downto 0);
  signal cpu0_gpio_out, cpu1_gpio_out : std_logic_vector(23 downto 0);
  signal cpu_gpio_oen                 : std_logic_vector(23 downto 0);
  signal cpu_gpio_out, cpu_gpio_in    : std_logic_vector(23 downto 0);

  signal VME_DATA_b_out    : std_logic_vector(31 downto 0);
  signal VME_ADDR_b_out    : std_logic_vector(31 downto 1);
  signal VME_LWORD_n_b_out : std_logic;
  signal VME_DATA_DIR_int  : std_logic;
  signal VME_ADDR_DIR_int  : std_logic;
  signal VME_BERR_n        : std_logic;
  signal VME_IRQ_n         : std_logic_vector(6 downto 0);

  signal mt_host_msg_irq  : std_logic;
  signal mt_debug_msg_irq : std_logic;
  signal vic_master_irq   : std_logic;

  signal powerup_reset_cnt : unsigned(7 downto 0) := "00000000";
  signal powerup_rst_n     : std_logic            := '0';
  signal sys_locked        : std_logic;

begin  -- architecture arch

  U_Buf_CLK_PLL : IBUFGDS
    generic map (
      DIFF_TERM    => TRUE,
      IBUF_LOW_PWR => TRUE)
    port map (
      O  => clk_125m_pllref,
      I  => clk_125m_pllref_p_i,
      IB => clk_125m_pllref_n_i);

  cmp_sys_clk_pll : PLL_BASE
    generic map (
      BANDWIDTH          => "OPTIMIZED",
      CLK_FEEDBACK       => "CLKFBOUT",
      COMPENSATION       => "INTERNAL",
      DIVCLK_DIVIDE      => 1,
      CLKFBOUT_MULT      => 8,
      CLKFBOUT_PHASE     => 0.000,
      CLKOUT0_DIVIDE     => 16,  -- 62.5 MHz
      CLKOUT0_PHASE      => 0.000,
      CLKOUT0_DUTY_CYCLE => 0.500,
      CLKIN_PERIOD       => 8.0,
      REF_JITTER         => 0.016)
    port map (
      CLKFBOUT => pllout_clk_fb_pllref,
      CLKOUT0  => pllout_clk_sys,
      LOCKED   => sys_locked,
      RST      => '0',
      CLKFBIN  => pllout_clk_fb_pllref,
      CLKIN    => clk_125m_pllref);

  cmp_clk_sys_buf : BUFG
    port map (
      O => clk_sys,
      I => pllout_clk_sys);

  p_powerup_reset : process(clk_sys)
  begin
    if rising_edge(clk_sys) then
      if(VME_RST_n_i = '0' or rst_n_a_i = '0') then
        powerup_rst_n <= '0';
      elsif sys_locked = '1' then
        if(powerup_reset_cnt = "11111111") then
          powerup_rst_n <= '1';
        else
          powerup_rst_n     <= '0';
          powerup_reset_cnt <= powerup_reset_cnt + 1;
        end if;
      else
        powerup_rst_n     <= '0';
        powerup_reset_cnt <= "00000000";
      end if;
    end if;
  end process;

  rst_n_sys <= powerup_rst_n;

  gen_with_vme64_core : if not g_sim_bypass_vme generate
    U_VME_Core : xvme64x_core
      generic map (
        g_CLOCK_PERIOD    => 16,
        g_DECODE_AM       => TRUE,
        g_USER_CSR_EXT    => FALSE,
        g_WB_GRANULARITY  => BYTE,
        g_MANUFACTURER_ID => c_CERN_ID,
        g_BOARD_ID        => c_SVEC_ID,
        g_REVISION_ID     => c_SVEC_REVISION_ID,
        g_PROGRAM_ID      => c_SVEC_PROGRAM_ID)
      port map (
        clk_i           => clk_sys,
        rst_n_i         => rst_n_sys,
        vme_i.as_n      => VME_AS_n_i,
        vme_i.rst_n     => rst_n_sys,
        vme_i.write_n   => VME_WRITE_n_i,
        vme_i.am        => VME_AM_i,
        vme_i.ds_n      => VME_DS_n_i,
        vme_i.ga        => VME_GA_i,
        vme_i.lword_n   => VME_LWORD_n_b,
        vme_i.addr      => VME_ADDR_b,
        vme_i.data      => VME_DATA_b,
        vme_i.iack_n    => VME_IACK_n_i,
        vme_i.iackin_n  => VME_IACKIN_n_i,
        vme_o.berr_n    => VME_BERR_n,
        vme_o.dtack_n   => VME_DTACK_n_o,
        vme_o.retry_n   => VME_RETRY_n_o,
        vme_o.retry_oe  => VME_RETRY_OE_o,
        vme_o.lword_n   => VME_LWORD_n_b_out,
        vme_o.data      => VME_DATA_b_out,
        vme_o.addr      => VME_ADDR_b_out,
        vme_o.irq_n     => VME_IRQ_n,
        vme_o.iackout_n => VME_IACKOUT_n_o,
        vme_o.dtack_oe  => VME_DTACK_OE_o,
        vme_o.data_dir  => VME_DATA_DIR_int,
        vme_o.data_oe_n => VME_DATA_OE_N_o,
        vme_o.addr_dir  => VME_ADDR_DIR_int,
        vme_o.addr_oe_n => VME_ADDR_OE_N_o,
        wb_o            => cnx_slave_in(c_MASTER_VME),
        wb_i            => cnx_slave_out(c_MASTER_VME),
        int_i           => vic_master_irq);

    VME_DATA_b    <= VME_DATA_b_out    when VME_DATA_DIR_int = '1' else (others => 'Z');
    VME_ADDR_b    <= VME_ADDR_b_out    when VME_ADDR_DIR_int = '1' else (others => 'Z');
    VME_LWORD_n_b <= VME_LWORD_n_b_out when VME_ADDR_DIR_int = '1' else 'Z';

    VME_ADDR_DIR_o <= VME_ADDR_DIR_int;
    VME_DATA_DIR_o <= VME_DATA_DIR_int;

    --  BERR and IRQ vme signals are inverted by the drivers. See SVEC schematics.
    VME_BERR_o  <= not VME_BERR_n;
    VME_IRQ_n_o <= not VME_IRQ_n;
  end generate gen_with_vme64_core;

  gen_without_vme64_core : if g_sim_bypass_vme generate
    -- synthesis translate_off
    cnx_slave_in(c_MASTER_VME) <= sim_wb_i;
    sim_wb_o                   <= cnx_slave_out(c_MASTER_VME);
  -- synthesis translate_on
  end generate gen_without_vme64_core;

  U_Intercon : xwb_sdb_crossbar
    generic map (
      g_num_masters => c_NUM_WB_SLAVES,
      g_num_slaves  => c_NUM_WB_MASTERS,
      g_registered  => TRUE,
      g_wraparound  => TRUE,
      g_layout      => c_INTERCONNECT_LAYOUT,
      g_sdb_addr    => c_SDB_ADDRESS)
    port map (
      clk_sys_i => clk_sys,
      rst_n_i   => rst_n_sys,
      slave_i   => cnx_slave_in,
      slave_o   => cnx_slave_out,
      master_i  => cnx_master_in,
      master_o  => cnx_master_out);

  U_VIC : xwb_vic
    generic map (
      g_interface_mode      => PIPELINED,
      g_address_granularity => BYTE,
      g_num_interrupts      => 2,
      g_init_vectors        => c_VIC_VECTOR_TABLE)
    port map (
      clk_sys_i    => clk_sys,
      rst_n_i      => rst_n_sys,
      slave_i      => cnx_master_out(c_SLAVE_VIC),
      slave_o      => cnx_master_in(c_SLAVE_VIC),
      irqs_i(0)    => mt_host_msg_irq,
      irqs_i(1)    => mt_debug_msg_irq,
      irq_master_o => vic_master_irq);

  U_Mock_Turtle : mock_turtle_core
    generic map (
      g_config            => c_mt_config,
      g_with_rmq          => FALSE,
      g_with_white_rabbit => FALSE)
    port map (
      clk_i           => clk_sys,
      rst_n_i         => rst_n_sys,
      dp_master_o     => dp_wb_out,
      dp_master_i     => dp_wb_in,
      host_slave_i    => cnx_master_out(c_SLAVE_MT),
      host_slave_o    => cnx_master_in(c_SLAVE_MT),
      host_irq_o      => mt_host_msg_irq,
      debug_msg_irq_o => mt_debug_msg_irq);

  U_GPIO_CPU0 : xwb_gpio_port
    generic map (
      g_interface_mode         => PIPELINED,
      g_address_granularity    => BYTE,
      g_num_pins               => 24,
      g_with_builtin_tristates => FALSE)
    port map (
      clk_sys_i  => clk_sys,
      rst_n_i    => rst_n_sys,
      slave_i    => dp_wb_out(0),
      slave_o    => dp_wb_in(0),
      gpio_out_o => cpu0_gpio_out,
      gpio_in_i  => cpu_gpio_in,
      gpio_oen_o => cpu0_gpio_oen);

  U_GPIO_CPU1 : xwb_gpio_port
    generic map (
      g_interface_mode         => PIPELINED,
      g_address_granularity    => BYTE,
      g_num_pins               => 24,
      g_with_builtin_tristates => FALSE)
    port map (
      clk_sys_i  => clk_sys,
      rst_n_i    => rst_n_sys,
      slave_i    => dp_wb_out(1),
      slave_o    => dp_wb_in(1),
      gpio_out_o => cpu1_gpio_out,
      gpio_in_i  => cpu_gpio_in,
      gpio_oen_o => cpu1_gpio_oen);

  cpu_gpio_out <= cpu0_gpio_out or cpu1_gpio_out;
  cpu_gpio_oen <= cpu0_gpio_oen or cpu1_gpio_oen;

  -- FP GPIO directions
  fp_gpio1_a2b_o  <= cpu_gpio_oen(0);
  fp_gpio2_a2b_o  <= cpu_gpio_oen(1);
  fp_gpio34_a2b_o <= cpu_gpio_oen(2);

  -- FP GPIO bidir in/out (3 and 4 share the same direction line)
  fp_gpio1_b <= cpu_gpio_out(0) when cpu_gpio_oen(0) = '1' else 'Z';
  fp_gpio2_b <= cpu_gpio_out(1) when cpu_gpio_oen(1) = '1' else 'Z';
  fp_gpio3_b <= cpu_gpio_out(2) when cpu_gpio_oen(2) = '1' else 'Z';
  fp_gpio4_b <= cpu_gpio_out(3) when cpu_gpio_oen(2) = '1' else 'Z';

  -- gpio inputs (same for both CPUs)
  cpu_gpio_in(0) <= fp_gpio1_b;
  cpu_gpio_in(1) <= fp_gpio2_b;
  cpu_gpio_in(2) <= fp_gpio3_b;
  cpu_gpio_in(3) <= fp_gpio4_b;

  cpu_gpio_in(23 downto 4) <= cpu_gpio_out(23 downto 4);

  U_LED_Controller : gc_bicolor_led_ctrl
    generic map(
      g_NB_COLUMN    => 4,
      g_NB_LINE      => 2,
      g_CLK_FREQ     => 62500000,  -- in Hz
      g_REFRESH_RATE => 250)       -- in Hz
    port map(
      rst_n_i => rst_n_sys,
      clk_i   => clk_sys,

      led_intensity_i => "1100100",  -- in %

      led_state_i => cpu_gpio_out(23 downto 8),

      column_o   => fp_led_column_o,
      line_o     => fp_led_line_o,
      line_oen_o => fp_led_line_oen_o);

end architecture arch;
