/*
 * Copyright (C) 2015-2016 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <mockturtle-framework.h>

/**
 * It sends the MQ content.
 * In order to save space the function *does not* perform any
 * argument validation. This means that you may crash the system
 * if you provide the wrong mq_idx.
 * @param[in] mq_idx MQ index within the declaration
 * @param[in] count number of bytes to send
 */
void trtl_rt_mq_send(unsigned int mq_idx, unsigned int count)
{
	mq_send(_app->mq[mq_idx].type, _app->mq[mq_idx].index, count);
	if (_app->mq[mq_idx].flags & TRTL_RT_MQ_FLAGS_CLAIM) {
		mq_claim(_app->mq[mq_idx].type, _app->mq[mq_idx].index);
	}
}


/**
 * It builds and it sends a message over MQ.
 * The vargs will be copied into the payload message.
 * Beware that, internally, it uses trtl_rt_mq_send().
 * @param[in] mq_idx MQ index within the declaration
 * @param[in] msg_id message identifier
 * @param[in] n number of vargs
 */
void trtl_rt_mq_send_uint32(unsigned int mq_idx, uint8_t msg_id,
			    unsigned int n, ...)
{
	struct trtl_proto_header hdr = {
		.rt_app_id = _app->version.rt_id,
		.msg_id = msg_id,
		.slot_io = _app->mq[mq_idx].index & 0xF,
		.seq = msg_seq++,
		.len = n,
		.flags = 0x0,
		.trans = 0x0,
		.time = 0x0,
	};
	void *buf = _app->mq[mq_idx].buf;
	uint32_t *payload = rt_proto_payload_get(buf);
	va_list ap;
	int i;

	va_start(ap, n);
	for (i = 0; i < hdr.len;++i)
		payload[i] = va_arg(ap, uint32_t);
	va_end(ap);

	rt_proto_header_set(buf, &hdr);
	trtl_rt_mq_send(mq_idx, (sizeof(struct trtl_proto_header) / 4) +
			hdr.len);
}


/**
 * It builds and it sends a message over MQ.
 * The buffer will be copied into the payload message.
 * Beware that, internally, it uses trtl_rt_mq_send().
 * @param[in] mq_idx MQ index within the declaration
 * @param[in] msg_id message identifier
 * @param[in] n buffer size in bytes
 * @param[in] data buffer to send
 */
void trtl_rt_mq_send_buf(unsigned int mq_idx, uint8_t msg_id,
			 unsigned int n, void *data)
{
	struct trtl_proto_header hdr = {
		.rt_app_id = _app->version.rt_id,
		.msg_id = msg_id,
		.slot_io = _app->mq[mq_idx].index & 0xF,
		.seq = msg_seq++,
		.len = n / 4,
		.flags = 0x0,
		.trans = 0x0,
		.time = 0x0,
	};
	void *buf = _app->mq[mq_idx].buf;

	rt_memcpy(rt_proto_payload_get(buf), data, n);

	rt_proto_header_set(buf, &hdr);
	trtl_rt_mq_send(mq_idx, (sizeof(struct trtl_proto_header) / 4) +
			hdr.len);
}
