#!/usr/bin/env python
import PyUAL
import getopt, sys
import time
import socket
import struct

gdb_port = 3000
flag_verbose = False
flag_keep = False
flag_cpu = 0

CSR_BASE = 0x2c000
CSR_APPID      = CSR_BASE + 0x00
CSR_RESET      = CSR_BASE + 0x04
CSR_ENABLE     = CSR_BASE + 0x08
CSR_UADDR      = CSR_BASE + 0x0c
CSR_CORESEL    = CSR_BASE + 0x10
CSR_CORECNT    = CSR_BASE + 0x14
CSR_COREMEM    = CSR_BASE + 0x18
CSR_UDATA      = CSR_BASE + 0x1c
CSR_DBG_MSG    = CSR_BASE + 0x20
CSR_DBG_POLL   = CSR_BASE + 0x24
CSR_DBG_STATUS = CSR_BASE + 0x30
CSR_DBG_FORCE  = CSR_BASE + 0x34
CSR_DBG_INSN_READY = CSR_BASE + 0x38
CSR_DBG_CPU0_INSN  = CSR_BASE + 0x3c
CSR_DBG_CPU1_INSN  = CSR_BASE + 0x40
CSR_DBG_MBX0       = CSR_BASE + 0x44
CSR_DBG_MBX1       = CSR_BASE + 0x48

def read_corenbr(csr):
    return csr.readl(CSR_CORECNT)

def cmd_status(csr):
    print ("Status:")
    print (" APP ID:          0x{0:08x}".format(csr.readl(CSR_APPID)))
    print (" RESET:           0x{0:08x}".format(csr.readl(CSR_RESET)))
    print (" ENABLE:          0x{0:08x}".format(csr.readl(CSR_ENABLE)))
    print (" UADDR:           0x{0:08x}".format(csr.readl(CSR_UADDR)))
    print (" CORESEL:         0x{0:08x}".format(csr.readl(CSR_CORESEL)))
    nbr_cores = read_corenbr(csr)
    print (" CORECNT:         0x{0:08x}".format(nbr_cores))
    print (" COREMEM:         0x{0:08x}".format(csr.readl(CSR_COREMEM)))
    print (" DBG_POLL:        0x{0:08x}".format(csr.readl(CSR_DBG_POLL)))
    print (" DBG_STATUS:      0x{0:08x}".format(csr.readl(CSR_DBG_STATUS)))
    print (" DBG_FORCE:       0x{0:08x}".format(csr.readl(CSR_DBG_FORCE)))
    print (" DBG_INSN_READY:  0x{0:08x}".format(csr.readl(CSR_DBG_INSN_READY)))
    print ("")
    for i in range(nbr_cores):
        csr.writel(CSR_CORESEL, i);
        memsz = csr.readl(CSR_COREMEM)
        print ("core {}:  mem {} kB".format(i, memsz / 1024))

def cpu_to_mbx(cpu):
    m = {0 : CSR_DBG_MBX0, 1: CSR_DBG_MBX1 }
    reg = m.get(cpu, None)
    if reg is None:
        raise AssertionError
    return reg

def read_mbx(csr, cpu):
    return csr.readl(cpu_to_mbx(cpu))

def write_mbx(csr, cpu, val):
    csr.writel(cpu_to_mbx(cpu), val)

def exec_insn(csr, cpu, insn):
    assert cpu == 0
    csr.writel(CSR_DBG_CPU0_INSN, insn)

def exec_reg_to_mbx(csr, cpu, reg):
    exec_insn(csr, cpu, 0x7d001073 + (reg << 15))

def exec_mbx_to_reg(csr, cpu, reg):
    exec_insn(csr, cpu, 0x7d002073 + (reg << 7))

def exec_nop(csr, cpu):
    exec_insn(csr, cpu, 0x00000013)

def swap32(v):
    return struct.unpack("<I", struct.pack(">I", v))[0]

def in_debug_mode(csr, cpu):
    poll = csr.readl(CSR_DBG_STATUS)
    return ((poll >> cpu) & 1) == 1

def force_debug(csr, cpu):
    if in_debug_mode(csr, cpu):
        return
    csr.writel(CSR_DBG_FORCE, 1 << cpu)
    while not in_debug_mode(csr, cpu):
        pass
    csr.writel(CSR_DBG_FORCE, 0)

def cmd_mbx(csr):
    nbr_cores = read_corenbr(csr)
    for i in range(nbr_cores):
        print (" MBX{}: 0x{:08x}".format(i, read_mbx(csr, i)))

def cmd_stop(csr):
    cpu = 0
    if in_debug_mode(csr, cpu):
        print ("CPU #{} already stopped".format(cpu))
    else:
        force_debug(csr, cpu)

def cmd_start(csr):
    cpu = 0
    if in_debug_mode(csr, cpu):
        exec_insn(csr, cpu, 0x00100073) # ebreak
    else:
        print ("CPU #{} already running".format(cpu))

def cmd_demo(csr):
    # Un-reset CPUs
    csr.writel(CSR_RESET, 0)
    # Disable force
    csr.writel(CSR_DBG_FORCE, 0)

    cpu = 0
    exec_insn(csr, cpu, 0x001005b7) # lui	a1,0x100
    exec_insn(csr, cpu, 0x01c5e593) # ori	a1,a1,28
    exec_insn(csr, cpu, 0x06800513) # li	a0,104
    exec_insn(csr, cpu, 0x00a58023) # sb	a0,0(a1)
    exec_reg_to_mbx(csr, cpu, 11)


def cmd_reset(csr):
    # Un-reset CPUs
    csr.writel(CSR_RESET, 0xff)
    csr.writel(CSR_DBG_FORCE, 0xff)
    time.sleep(0.01)
    csr.writel(CSR_RESET, 0x00)
    csr.writel(CSR_DBG_FORCE, 0x00)


def cmd_term(csr):
    cpu = flag_cpu
    cpu_mask = 1 << cpu
    csr.writel(CSR_CORESEL, cpu);
    timeout = 1
    try:
        while True:
            poll = csr.readl(CSR_DBG_POLL)
            if poll & cpu_mask:
                c = csr.readl(CSR_DBG_MSG)
                sys.stdout.write(chr(c & 0xff))
                sys.stdout.flush()
                timeout = 1
            else:
                time.sleep(timeout / 1000.0)
                if timeout < 10:
                    timeout += 1
                elif timeout < 100:
                    timeout += 10
    except KeyboardInterrupt:
        sys.stdout.write("INT\n")
        pass
    except:
        raise
            

class GdbServer(object):
    def __init__(self, csr, cpu, port):
        self.csr = csr
        self.cpu = cpu
        self.port = port
        self.conn = None

    def checksum(self, packet):
        chk = 0
        for b in packet:
            chk += ord(b)
        return chk & 0xff

    def in_debug_mode(self):
        return in_debug_mode(self.csr, self.cpu)

    def force_debug(self):
        if self.in_debug_mode():
            return
        self.csr.writel(CSR_DBG_FORCE, 1 << self.cpu)
        while not self.in_debug_mode():
            pass
        self.csr.writel(CSR_DBG_FORCE, 0)

    def send(self, packet):
        pkt = "${}#{:02x}".format(packet, self.checksum(packet))
        if flag_verbose:
            print "> {}".format(packet)
        self.conn.send(pkt)

    def read_reg(self, reg):
        assert reg >= 0 and reg < 32
        exec_reg_to_mbx(self.csr, self.cpu, reg)
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        return read_mbx(self.csr, self.cpu)

    def write_reg(self, reg, val):
        assert reg >= 0 and reg < 32
        write_mbx(self.csr, self.cpu, val)
        exec_mbx_to_reg(self.csr, self.cpu, reg)

    def read_pc_via_ra(self):
        # Read current pc value but destroy ra.
        exec_insn(self.csr, self.cpu, 0x000000ef) # ra = pc + 4
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        exec_reg_to_mbx(self.csr, self.cpu, 1)
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        return read_mbx(self.csr, self.cpu) - 4

    def write_pc_via_ra(self):
        exec_insn(self.csr, self.cpu, 0x00008067) # ret
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)

    def advance_pc_4(self):
        exec_insn(self.csr, self.cpu, 0x00000263) # beqz zero, +4
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)

    def handle_c(self, packet):
        if len(packet) > 1:
            addr = int(packet[1:], 16)
            self.write_pc(addr)
        exec_insn(self.csr, self.cpu, 0x00100073) # ebreak
        # Wait until the CPU is in debug mode
        timeout = 1
        while True:
            if self.in_debug_mode():
                # Check twice due to possible race if the ebreak is not yet
                # executed
                if self.in_debug_mode():
                    break
            poll = self.csr.readl(CSR_DBG_POLL)
            if poll & (1 << self.cpu):
                c = self.csr.readl(CSR_DBG_MSG)
                sys.stdout.write(chr(c & 0xff))
                sys.stdout.flush()
                timeout = 1
            # FIXME: not reliable
            try:
                time.sleep(timeout / 1000.0)
            except KeyboardInterrupt:
                self.force_debug()
                self.send('S02')
                return
            if timeout < 10:
                timeout += 1
            elif timeout < 100:
                timeout += 10
        self.send('S05')
            
    def handle_g(self, packet):
        # Serious work: read all registers.
        regs = [None] * 32
        pc = None
        res = ''
        for r in range(32):
            regs[r] = self.read_reg(r)
            res += "{:08x}".format(swap32(regs[r]))
        # Read PC
        pc = self.read_pc_via_ra()
        res += "{:08x}".format(swap32(pc))
        self.write_reg(1, regs[1])
        self.send(res)

    def handle_G(self, packet):
        if len(packet) != 1 + 33 * 8:
            self.send('')
            return
        regs = [None] * 33
        for r in range(33):
            regs[r] = swap32(int(packet[r*8 + 1: r*8 + 9], 16))
        self.write_reg(1, regs[32])
        self.write_pc_via_ra()
        for r in range(32):
            self.write_reg(r, regs[r])
        self.send('OK')

    def handle_H(self, packet):
        if packet == 'Hg0':
            self.send('OK')
        else:
            self.send('')

    def handle_k(self, packet):
        # kill
        pass

    def handle_M(self, packet):
        # Write data to memory.
        comma = packet.find(',')
        colon = packet.find(':')
        if comma == -1 or colon == -1 or colon < comma:
            self.send('')
            return
        addr = int(packet[1:comma], 16)
        nbytes = int(packet[comma + 1:colon], 16)
        data = packet[colon + 1:]
        if len(data) != nbytes * 2:
            self.send('E01')
            return
        # Save a0, a1
        a0 = self.read_reg(10)
        a1 = self.read_reg(11)
        self.write_reg(10, addr)
        if (addr % 4) == 0:
            while nbytes >= 4:
                w = swap32(int(data[:8], 16))
                data = data[8:]
                nbytes -= 4
                self.write_reg(11, w)
                exec_insn(self.csr, self.cpu, 0x00b52023) # sw a1,0(a0)
                exec_insn(self.csr, self.cpu, 0x00450513) # addi a0, a0, 4
        while nbytes > 0:
            b = int(data[:2], 16)
            data = data[2:]
            nbytes -= 1
            self.write_reg(11, b)
            exec_insn(self.csr, self.cpu, 0x00b50023) # sb a1,0(a0)
            exec_insn(self.csr, self.cpu, 0x00150513) # addi a0, a0, 1
        self.write_reg(10, a0)
        self.write_reg(11, a1)
        self.send('OK')

    def handle_m(self, packet):
        # Read data from memory.
        comma = packet.find(',')
        if comma == -1:
            self.send('')
            return
        addr = int(packet[1:comma], 16)
        nbytes = int(packet[comma + 1:], 16)
        res = ''
        # Save a0, a1
        a0 = self.read_reg(10)
        a1 = self.read_reg(11)
        self.write_reg(10, addr)
        while nbytes > 0:
            exec_insn(self.csr, self.cpu, 0x00054583) # lbu a1,0(a0)
            exec_insn(self.csr, self.cpu, 0x00150513) # addi a0, a0, 1
            b = self.read_reg(11)
            res += '{:02x}'.format(b)
            nbytes -= 1
        self.write_reg(10, a0)
        self.write_reg(11, a1)
        self.send(res)

    def handle_p(self, packet):
        # Read a specific register
        regnum = int(packet[1:], 16)
        if regnum == (65 + 0x301):
            # MISA CSR
            self.send('{:08x}'.format((1 << 30) | (1 << (ord('I') - 65))))
        else:
            self.send('E01')

    def handle_P(self, packet):
        # Write a specific register
        self.send('')

    def handle_q_supported(self, packet):
        self.send('')

    def handle_q(self, packet):
        if packet.startswith('qSupported:'):
            self.handle_q_supported(packet)
        else:
            self.send('')

    def handle_s(self, packet):
        # Single step
        ra = self.read_reg(1)
        if len(packet) > 1:
            pc = int(packet[1:], 16)
            self.write_pc(pc)
        else:
            pc = self.read_pc_via_ra()
        # Read insn
        self.write_reg(1, pc)
        exec_insn(self.csr, self.cpu, 0x0000a083) # lw ra,0(ra)
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        insn = self.read_reg(1)
        self.write_reg(1, ra)
        if flag_verbose:
            print ("execute: {:08x} at pc={:08x}".format(insn, pc))
        exec_insn(self.csr, self.cpu, insn)
        exec_nop(self.csr, self.cpu)
        exec_nop(self.csr, self.cpu)
        if (insn & 0x77) == 0x67:
            # JAL/JLR: pc is always updated
            pass
        elif (insn & 0x7f) == 0x63:
            # A branch insn
            ra = self.read_reg(1)
            npc = self.read_pc_via_ra()
            self.write_reg(1, ra)
            if npc != pc:
                self.advance_pc_4()
        else:
            self.advance_pc_4()
        self.send('S05')

    def handle_vCont(self, packet):
        if packet == '?':
            self.send('vCont;c;s')
        else:
            self.send('')

    def handle_v(self, packet):
        if packet.startswith('vMustReplyEmpty:'):
            self.send('')
        elif packet.startswith('vCont'):
            self.handle_vCont(packet[5:])
        else:
            self.send('')

    def handle_X(self, packet):
        # Write binary data to memory.
        self.send('')

    def handle_qm(self, packet):
        self.send('S05')

    def gdb_command(self, packet):
        if flag_verbose:
            print "< {}".format(packet)
        cmd = packet[0]
        m = getattr(self, 'handle_' + cmd, None)
        if m:
            m(packet)
        elif cmd == '?':
            self.handle_qm(packet)
        else:
            if flag_verbose:
                print ('Unknown command: {}'.format(cmd))
            self.send('');

    def run_server(self, sock):
        sys.stdout.write("Waiting for connection on port {}\n".format(gdb_port))
        self.conn, client_addr = sock.accept()
        sys.stdout.write("connection from {}\n".format(client_addr))
        while True:
            # Wait for a packet
            packet = ""
            data = self.conn.recv(2048)
            # Exit in case of disconnection
            if len(data) == 0:
                break
            for c in data:
                if len(packet) == 0:
                    if c == '$':
                        packet = c
                    else:
                        pass
                else:
                    packet += c
                    if len(packet) > 3 and packet[-3] == '#':
                        chk = self.checksum(packet[1:-3])
                        pchk = int(packet[-2:], 16)
                        if chk != pchk:
                            print 'chkerr: {:x} vs {:x}'.format(chk, pchk)
                            # print 'pkt: {}'.format(packet[1:-3])
                            raise Exception('invalid checksum')
                        self.conn.send('+')
                        packet = packet[1:-3]
                        if packet == '':
                            pass
                        else:
                            self.gdb_command(packet)
                            packet = ''

    def run(self):
        self.force_debug()
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(('', gdb_port))
        sock.listen(1)
        self.run_server(sock)
        while flag_keep:
            self.run_server(sock)


def cmd_gdb(csr):
    # Clear reset
    csr.writel(CSR_RESET, 0)
    # Disable force
    csr.writel(CSR_DBG_FORCE, 0)

    gdb = GdbServer(csr, 0, gdb_port)
    gdb.run()


def main():
    global flag_verbose, flag_keep, flag_cpu
    slot = None
    csr = None
    try:
        opts, args = getopt.getopt(sys.argv[1:], "s:vkc:")
    except getopt.GetoptError as err:
        print str(err)
        sys.exit(2)
    for o, a in opts:
        if o == "-s":
            slot = int(a)
        elif o == "-v":
            flag_verbose = True
        elif o == "-k":
            flag_keep = True
        elif o == "-c":
            flag_cpu = int(a)
        else:
            assert False, "unhandled option"
    
    if slot:
        mt_desc = PyUAL.PyUALVME(4, 0x39, 0x80000, slot << 19, 1)
        csr = PyUAL.PyUAL("vme", mt_desc)
    else:
        print "missing slot (try -s SLOT)"
        sys.exit(2)

    cmds = {'status': cmd_status,
            'demo':   cmd_demo,
            'mbx':    cmd_mbx,
            'term':   cmd_term,
            'reset':  cmd_reset,
            'stop':   cmd_stop,
            'start':  cmd_start,
            'gdb':    cmd_gdb}
    for cmd in args:
        f = cmds.get(cmd, None)
        if f:
            f(csr)
        else:
            print "Unknown command {}".format(cmd)
            sys.exit(2)

if __name__ == "__main__":
    main()

